@@ -21,10 +21,14 @@ local file = require("apisix.cli.file")
 local schema = require("apisix.cli.schema")
 local ngx_tpl = require("apisix.cli.ngx_tpl")
 local cli_ip = require("apisix.cli.ip")
+local snippet = require("apisix.cli.snippet")
 local profile = require("apisix.core.profile")
 local template = require("resty.template")
 local argparse = require("argparse")
 local pl_path = require("pl.path")
+local lfs = require("lfs")
+local signal = require("posix.signal")
+local errno = require("posix.errno")
 
 local stderr = io.stderr
 local ipairs = ipairs
@@ -36,7 +40,9 @@ local tonumber = tonumber
 local io_open = io.open
 local execute = os.execute
 local os_rename = os.rename
+local os_remove = os.remove
 local table_insert = table.insert
+local table_remove = table.remove
 local getenv = os.getenv
 local max = math.max
 local floor = math.floor
@@ -60,6 +66,7 @@ stop:       stop the apisix server
 quit:       stop the apisix server gracefully
 restart:    restart the apisix server
 reload:     reload the apisix server
+test:       test the generated nginx.conf
 version:    print the version of apisix
 ]])
 end
@@ -239,12 +246,9 @@ Please modify "admin_key" in conf/config.yaml .
     end
 
     local or_info = util.execute_cmd("openresty -V 2>&1")
-    local with_module_status = true
     if or_info and not or_info:find("http_stub_status_module", 1, true) then
-        stderr:write("'http_stub_status_module' module is missing in ",
-                     "your openresty, please check it out. Without this ",
-                     "module, there will be fewer monitoring indicators.\n")
-        with_module_status = false
+        util.die("'http_stub_status_module' module is missing in ",
+                 "your openresty, please check it out.\n")
     end
 
     local use_apisix_openresty = true
@@ -252,6 +256,13 @@ Please modify "admin_key" in conf/config.yaml .
         use_apisix_openresty = false
     end
 
+    local enable_http = true
+    if not yaml_conf.apisix.enable_admin and yaml_conf.apisix.stream_proxy and
+        yaml_conf.apisix.stream_proxy.only ~= false
+    then
+        enable_http = false
+    end
+
     local enabled_discoveries = {}
     for name in pairs(yaml_conf.discovery or {}) do
         enabled_discoveries[name] = true
@@ -341,6 +352,10 @@ Please modify "admin_key" in conf/config.yaml .
         end
     end
 
+    if enabled_stream_plugins["prometheus"] and not prometheus_server_addr then
+        util.die("L4 prometheus metric should be exposed via export server\n")
+    end
+
     local ip_port_to_check = {}
 
     local function listen_table_insert(listen_table, scheme, ip, port, enable_http2, enable_ipv6)
@@ -469,9 +484,8 @@ Please modify "admin_key" in conf/config.yaml .
         -- Therefore we need to check the absolute version instead
         cert_path = pl_path.abspath(cert_path)
 
-        local ok, err = util.is_file_exist(cert_path)
-        if not ok then
-            util.die(err, "\n")
+        if not pl_path.exists(cert_path) then
+            util.die("certificate path", cert_path, "doesn't exist\n")
         end
 
         yaml_conf.apisix.ssl.ssl_trusted_certificate = cert_path
@@ -526,6 +540,15 @@ Please modify "admin_key" in conf/config.yaml .
         proxy_mirror_timeouts = yaml_conf.plugin_attr["proxy-mirror"].timeout
     end
 
+    local conf_server, err = snippet.generate_conf_server(env, yaml_conf)
+    if err then
+        util.die(err, "\n")
+    end
+
+    if yaml_conf.deployment and yaml_conf.deployment.role then
+        env.deployment_role = yaml_conf.deployment.role
+    end
+
     -- Using template.render
     local sys_conf = {
         use_openresty_1_17 = use_openresty_1_17,
@@ -533,9 +556,9 @@ Please modify "admin_key" in conf/config.yaml .
         lua_cpath = env.pkg_cpath_org,
         os_name = util.trim(util.execute_cmd("uname")),
         apisix_lua_home = env.apisix_home,
-        with_module_status = with_module_status,
         use_apisix_openresty = use_apisix_openresty,
         error_log = {level = "warn"},
+        enable_http = enable_http,
         enabled_discoveries = enabled_discoveries,
         enabled_plugins = enabled_plugins,
         enabled_stream_plugins = enabled_stream_plugins,
@@ -545,6 +568,7 @@ Please modify "admin_key" in conf/config.yaml .
         control_server_addr = control_server_addr,
         prometheus_server_addr = prometheus_server_addr,
         proxy_mirror_timeouts = proxy_mirror_timeouts,
+        conf_server = conf_server,
     }
 
     if not yaml_conf.apisix then
@@ -611,6 +635,14 @@ Please modify "admin_key" in conf/config.yaml .
                 sys_conf["dns_resolver"][i] = "[" .. r .. "]"
             end
         end
+
+        -- check if the dns_resolver is ipv6 address with zone_id
+        -- Nginx does not support this form
+        if r:find("%%") then
+            stderr:write("unsupported DNS resolver: " .. r ..
+                         ", would ignore this item\n")
+            table_remove(sys_conf["dns_resolver"], i)
+        end
     end
 
     local env_worker_processes = getenv("APISIX_WORKER_PROCESSES")
@@ -731,15 +763,20 @@ local function start(env, ...)
     local pid = util.read_file(pid_path)
     pid = tonumber(pid)
     if pid then
-        local lsof_cmd = "lsof -p " .. pid
-        local res, err = util.execute_cmd(lsof_cmd)
-        if not (res and res == "") then
-            if not res then
-                print(err)
-            else
-                print("APISIX is running...")
-            end
+        if pid <= 0 then
+            print("invalid pid")
+            return
+        end
 
+        local signone = 0
+
+        local ok, err, err_no = signal.kill(pid, signone)
+        if ok then
+            print("APISIX is running...")
+            return
+        -- no such process
+        elseif err_no ~= errno.ESRCH then
+            print(err)
             return
         end
 
@@ -758,23 +795,29 @@ local function start(env, ...)
     if customized_yaml then
         profile.apisix_home = env.apisix_home .. "/"
         local local_conf_path = profile:yaml_path("config")
+        local local_conf_path_bak = local_conf_path .. ".bak"
 
-        local err = util.execute_cmd_with_error("mv " .. local_conf_path .. " "
-                                                .. local_conf_path .. ".bak")
-        if #err > 0 then
-            util.die("failed to mv config to backup, error: ", err)
+        local ok, err = os_rename(local_conf_path, local_conf_path_bak)
+        if not ok then
+            util.die("failed to backup config, error: ", err)
         end
-        err = util.execute_cmd_with_error("ln " .. customized_yaml .. " " .. local_conf_path)
-        if #err > 0 then
-            util.execute_cmd("mv " .. local_conf_path .. ".bak " .. local_conf_path)
-            util.die("failed to link customized config, error: ", err)
+        local ok, err1 = lfs.link(customized_yaml, local_conf_path)
+        if not ok then
+            ok, err = os_rename(local_conf_path_bak,  local_conf_path)
+            if not ok then
+                util.die("failed to recover original config file, error: ", err)
+            end
+            util.die("failed to link customized config, error: ", err1)
         end
 
         print("Use customized yaml: ", customized_yaml)
     end
 
     init(env)
-    init_etcd(env, args)
+
+    if env.deployment_role ~= "data_plane" then
+        init_etcd(env, args)
+    end
 
     util.execute_cmd(env.openresty_args)
 end
@@ -782,15 +825,15 @@ end
 
 local function cleanup()
     local local_conf_path = profile:yaml_path("config")
-    local bak_exist = io_open(local_conf_path .. ".bak")
-    if bak_exist then
-        local err = util.execute_cmd_with_error("rm " .. local_conf_path)
-        if #err > 0 then
+    local local_conf_path_bak = local_conf_path .. ".bak"
+    if pl_path.exists(local_conf_path_bak) then
+        local ok, err = os_remove(local_conf_path)
+        if not ok then
             print("failed to remove customized config, error: ", err)
         end
-        err = util.execute_cmd_with_error("mv " .. local_conf_path .. ".bak " .. local_conf_path)
-        if #err > 0 then
-            util.die("failed to mv original config file, error: ", err)
+        ok, err = os_rename(local_conf_path_bak,  local_conf_path)
+        if not ok then
+            util.die("failed to recover original config file, error: ", err)
         end
     end
 end
@@ -799,9 +842,10 @@ end
 local function test(env, backup_ngx_conf)
     -- backup nginx.conf
     local ngx_conf_path = env.apisix_home .. "/conf/nginx.conf"
-    local ngx_conf_exist = util.is_file_exist(ngx_conf_path)
+    local ngx_conf_path_bak = ngx_conf_path .. ".bak"
+    local ngx_conf_exist = pl_path.exists(ngx_conf_path)
     if ngx_conf_exist then
-        local ok, err = os_rename(ngx_conf_path, ngx_conf_path .. ".bak")
+        local ok, err = os_rename(ngx_conf_path, ngx_conf_path_bak)
         if not ok then
             util.die("failed to backup nginx.conf, error: ", err)
         end
@@ -815,7 +859,7 @@ local function test(env, backup_ngx_conf)
 
     -- restore nginx.conf
     if ngx_conf_exist then
-        local ok, err = os_rename(ngx_conf_path .. ".bak", ngx_conf_path)
+        local ok, err = os_rename(ngx_conf_path_bak, ngx_conf_path)
         if not ok then
             util.die("failed to restore original nginx.conf, error: ", err)
         end
