@@ -19,6 +19,7 @@ local core          = require("apisix.core")
 local config_util   = require("apisix.core.config_util")
 local enable_debug  = require("apisix.debug").enable_debug
 local wasm          = require("apisix.wasm")
+local expr          = require("resty.expr.v1")
 local ngx           = ngx
 local crc32         = ngx.crc32_short
 local ngx_exit      = ngx.exit
@@ -40,8 +41,11 @@ local stream_local_plugins_hash = core.table.new(0, 32)
 local merged_route = core.lrucache.new({
     ttl = 300, count = 512
 })
+local expr_lrucache = core.lrucache.new({
+    ttl = 300, count = 512
+})
 local local_conf
-
+local check_plugin_metadata
 
 local _M = {
     version         = 0.3,
@@ -68,6 +72,10 @@ local function sort_plugin(l, r)
     return l.priority > r.priority
 end
 
+local function custom_sort_plugin(l, r)
+    return l._meta.priority > r._meta.priority
+end
+
 
 local PLUGIN_TYPE_HTTP = 1
 local PLUGIN_TYPE_STREAM = 2
@@ -142,6 +150,19 @@ local function load_plugin(name, plugins_list, plugin_type)
         end
 
         properties.disable = plugin_injected_schema.disable
+
+        if properties._meta then
+            core.log.error("invalid plugin [", name,
+                           "]: found forbidden '_meta' field in the schema")
+            return
+        end
+
+        properties._meta = plugin_injected_schema._meta
+        -- new injected fields should be added under `_meta`
+        -- 1. so we won't break user's code when adding any new injected fields
+        -- 2. the semantics is clear, especially in the doc and in the caller side
+        -- TODO: move the `disable` to `_meta` too
+
         plugin.schema['$comment'] = plugin_injected_schema['$comment']
     end
 
@@ -255,7 +276,7 @@ local function load_stream(plugin_names)
 end
 
 
-function _M.load(config)
+local function get_plugin_names(config)
     local http_plugin_names
     local stream_plugin_names
 
@@ -277,7 +298,7 @@ function _M.load(config)
         local plugins_conf = config.value
         -- plugins_conf can be nil when another instance writes into etcd key "/apisix/plugins/"
         if not plugins_conf then
-            return local_plugins
+            return true
         end
 
         for _, conf in ipairs(plugins_conf) do
@@ -289,6 +310,16 @@ function _M.load(config)
         end
     end
 
+    return false, http_plugin_names, stream_plugin_names
+end
+
+
+function _M.load(config)
+    local ignored, http_plugin_names, stream_plugin_names = get_plugin_names(config)
+    if ignored then
+        return local_plugins
+    end
+
     if ngx.config.subsystem == "http" then
         if not http_plugin_names then
             core.log.error("failed to read plugin list from local file")
@@ -354,8 +385,34 @@ local function trace_plugins_info_for_debug(ctx, plugins)
     end
 end
 
+local function meta_filter(ctx, plugin_name, plugin_conf)
+    local filter = plugin_conf._meta and plugin_conf._meta.filter
+    if not filter then
+        return true
+    end
 
-function _M.filter(ctx, conf, plugins, route_conf)
+    local ex, ok, err
+    if ctx then
+        ex, err = expr_lrucache(plugin_name .. ctx.conf_type .. ctx.conf_id,
+                                 ctx.conf_version, expr.new, filter)
+    else
+        ex, err = expr.new(filter)
+    end
+    if not ex then
+        core.log.warn("failed to get the 'vars' expression: ", err ,
+                         " plugin_name: ", plugin_name)
+        return true
+    end
+    ok, err = ex:eval()
+    if err then
+        core.log.warn("failed to run the 'vars' expression: ", err,
+                         " plugin_name: ", plugin_name)
+        return true
+    end
+    return ok
+end
+
+function _M.filter(ctx, conf, plugins, route_conf, phase)
     local user_plugin_conf = conf.value.plugins
     if user_plugin_conf == nil or
        core.table.nkeys(user_plugin_conf) == 0 then
@@ -365,13 +422,19 @@ function _M.filter(ctx, conf, plugins, route_conf)
         return plugins or core.tablepool.fetch("plugins", 0, 0)
     end
 
+    local custom_sort = false
     local route_plugin_conf = route_conf and route_conf.value.plugins
     plugins = plugins or core.tablepool.fetch("plugins", 32, 0)
     for _, plugin_obj in ipairs(local_plugins) do
         local name = plugin_obj.name
         local plugin_conf = user_plugin_conf[name]
 
-        if type(plugin_conf) == "table" and not plugin_conf.disable then
+        if type(plugin_conf) ~= "table" then
+            goto continue
+        end
+
+        local matched = meta_filter(ctx, name, plugin_conf)
+        if not plugin_conf.disable and matched then
             if plugin_obj.run_policy == "prefer_route" and route_plugin_conf ~= nil then
                 local plugin_conf_in_route = route_plugin_conf[name]
                 if plugin_conf_in_route and not plugin_conf_in_route.disable then
@@ -379,15 +442,63 @@ function _M.filter(ctx, conf, plugins, route_conf)
                 end
             end
 
+            if plugin_conf._meta and plugin_conf._meta.priority then
+                custom_sort = true
+            end
             core.table.insert(plugins, plugin_obj)
             core.table.insert(plugins, plugin_conf)
-
-            ::continue::
         end
+
+        ::continue::
     end
 
     trace_plugins_info_for_debug(ctx, plugins)
 
+    if custom_sort then
+        local tmp_plugin_objs = core.tablepool.fetch("tmp_plugin_objs", 0, #plugins / 2)
+        local tmp_plugin_confs = core.tablepool.fetch("tmp_plugin_confs", #plugins / 2, 0)
+
+        for i = 1, #plugins, 2 do
+            local plugin_obj = plugins[i]
+            local plugin_conf = plugins[i + 1]
+
+            -- in the rewrite phase, the plugin executes in the following order:
+            -- 1. execute the rewrite phase of the plugins on route(including the auth plugins)
+            -- 2. merge plugins from consumer and route
+            -- 3. execute the rewrite phase of the plugins on consumer(phase: rewrite_in_consumer)
+            -- in this case, we need to skip the plugins that was already executed(step 1)
+            if phase == "rewrite_in_consumer" and not plugin_conf._from_consumer then
+                plugin_conf._skip_rewrite_in_consumer = true
+            end
+
+            tmp_plugin_objs[plugin_conf] = plugin_obj
+            core.table.insert(tmp_plugin_confs, plugin_conf)
+
+            if not plugin_conf._meta then
+                plugin_conf._meta = core.table.new(0, 1)
+                plugin_conf._meta.priority = plugin_obj.priority
+            else
+                if not plugin_conf._meta.priority then
+                    plugin_conf._meta.priority = plugin_obj.priority
+                end
+            end
+        end
+
+        sort_tab(tmp_plugin_confs, custom_sort_plugin)
+
+        local index
+        for i = 1, #tmp_plugin_confs do
+            index = i * 2 - 1
+            local plugin_conf = tmp_plugin_confs[i]
+            local plugin_obj = tmp_plugin_objs[plugin_conf]
+            plugins[index] = plugin_obj
+            plugins[index + 1] = plugin_conf
+        end
+
+        core.tablepool.release("tmp_plugin_objs", tmp_plugin_objs)
+        core.tablepool.release("tmp_plugin_confs", tmp_plugin_confs)
+    end
+
     return plugins
 end
 
@@ -554,19 +665,30 @@ end
 
 
 function _M.init_worker()
-    _M.load()
+    local _, http_plugin_names, stream_plugin_names = get_plugin_names()
 
     -- some plugins need to be initialized in init* phases
-    if ngx.config.subsystem == "http" and local_plugins_hash["prometheus"] then
-        require("apisix.plugins.prometheus.exporter").init()
+    if is_http and core.table.array_find(http_plugin_names, "prometheus") then
+        local prometheus_enabled_in_stream =
+            core.table.array_find(stream_plugin_names, "prometheus")
+        require("apisix.plugins.prometheus.exporter").http_init(prometheus_enabled_in_stream)
+    elseif not is_http and core.table.array_find(stream_plugin_names, "prometheus") then
+        require("apisix.plugins.prometheus.exporter").stream_init()
     end
 
+    -- someone's plugin needs to be initialized after prometheus
+    -- see https://github.com/apache/apisix/issues/3286
+    _M.load()
+
     if local_conf and not local_conf.apisix.enable_admin then
         init_plugins_syncer()
     end
 
     local plugin_metadatas, err = core.config.new("/plugin_metadata",
-        {automatic = true}
+        {
+            automatic = true,
+            checker = check_plugin_metadata
+        }
     )
     if not plugin_metadatas then
         error("failed to create etcd instance for fetching /plugin_metadatas : "
@@ -620,39 +742,62 @@ function _M.conf_version(conf)
 end
 
 
+local function check_single_plugin_schema(name, plugin_conf, schema_type, skip_disabled_plugin)
+    core.log.info("check plugin schema, name: ", name, ", configurations: ",
+        core.json.delay_encode(plugin_conf, true))
+    if type(plugin_conf) ~= "table" then
+        return false, "invalid plugin conf " ..
+            core.json.encode(plugin_conf, true) ..
+            " for plugin [" .. name .. "]"
+    end
+
+    local plugin_obj = local_plugins_hash[name]
+    if not plugin_obj then
+        if skip_disabled_plugin then
+            return true
+        else
+            return false, "unknown plugin [" .. name .. "]"
+        end
+    end
+
+    if plugin_obj.check_schema then
+        local disable = plugin_conf.disable
+        plugin_conf.disable = nil
+
+        local ok, err = plugin_obj.check_schema(plugin_conf, schema_type)
+        if not ok then
+            return false, "failed to check the configuration of plugin "
+                .. name .. " err: " .. err
+        end
+
+        if plugin_conf._meta and plugin_conf._meta.filter then
+            ok, err = expr.new(plugin_conf._meta.filter)
+            if not ok then
+                return nil, "failed to validate the 'vars' expression: " .. err
+            end
+        end
+
+        plugin_conf.disable = disable
+    end
+
+    return true
+end
+
+
+check_plugin_metadata = function(item)
+    return check_single_plugin_schema(item.id, item,
+        core.schema.TYPE_METADATA, true)
+end
+
+
+
 local function check_schema(plugins_conf, schema_type, skip_disabled_plugin)
     for name, plugin_conf in pairs(plugins_conf) do
-        core.log.info("check plugin schema, name: ", name, ", configurations: ",
-                      core.json.delay_encode(plugin_conf, true))
-        if type(plugin_conf) ~= "table" then
-            return false, "invalid plugin conf " ..
-                core.json.encode(plugin_conf, true) ..
-                " for plugin [" .. name .. "]"
+        local ok, err = check_single_plugin_schema(name, plugin_conf,
+            schema_type, skip_disabled_plugin)
+        if not ok then
+            return false, err
         end
-
-        local plugin_obj = local_plugins_hash[name]
-        if not plugin_obj then
-            if skip_disabled_plugin then
-                goto CONTINUE
-            else
-                return false, "unknown plugin [" .. name .. "]"
-            end
-        end
-
-        if plugin_obj.check_schema then
-            local disable = plugin_conf.disable
-            plugin_conf.disable = nil
-
-            local ok, err = plugin_obj.check_schema(plugin_conf, schema_type)
-            if not ok then
-                return false, "failed to check the configuration of plugin "
-                              .. name .. " err: " .. err
-            end
-
-            plugin_conf.disable = disable
-        end
-
-        ::CONTINUE::
     end
 
     return true
@@ -741,13 +886,26 @@ function _M.run_plugin(phase, plugins, api_ctx)
                 phase = "rewrite"
             end
             local phase_func = plugins[i][phase]
+
+            if phase == "rewrite" and plugins[i + 1]._skip_rewrite_in_consumer then
+                goto CONTINUE
+            end
+
             if phase_func then
                 plugin_run = true
-                local code, body = phase_func(plugins[i + 1], api_ctx)
+                local conf = plugins[i + 1]
+                local code, body = phase_func(conf, api_ctx)
                 if code or body then
                     if is_http then
                         if code >= 400 then
                             core.log.warn(plugins[i].name, " exits with http status code ", code)
+
+                            if conf._meta and conf._meta.error_response then
+                                -- Whether or not the original error message is output,
+                                -- always return the configured message
+                                -- so the caller can't guess the real error
+                                body = conf._meta.error_response
+                            end
                         end
 
                         core.response.exit(code, body)
@@ -760,6 +918,8 @@ function _M.run_plugin(phase, plugins, api_ctx)
                     end
                 end
             end
+
+            ::CONTINUE::
         end
         return api_ctx, plugin_run
     end
