@@ -404,10 +404,29 @@ local upstream_schema = {
         tls = {
             type = "object",
             properties = {
+                client_cert_id = id_schema,
                 client_cert = certificate_scheme,
                 client_key = private_key_schema,
+                verify = {
+                    type = "boolean",
+                    description = "Turn on server certificate verification, "..
+                        "currently only kafka upstream is supported",
+                    default = false,
+                },
             },
-            required = {"client_cert", "client_key"},
+            dependencies = {
+                client_cert = {
+                    required = {"client_key"},
+                    ["not"] = {required = {"client_cert_id"}}
+                },
+                client_key = {
+                    required = {"client_cert"},
+                    ["not"] = {required = {"client_cert_id"}}
+                },
+                client_cert_id = {
+                    ["not"] = {required = {"client_client", "client_key"}}
+                }
+            }
         },
         keepalive_pool = {
             type = "object",
@@ -495,7 +514,7 @@ local upstream_schema = {
     oneOf = {
         {required = {"type", "nodes"}},
         {required = {"type", "service_name", "discovery_type"}},
-    },
+    }
 }
 
 -- TODO: add more nginx variable support
@@ -713,6 +732,14 @@ _M.ssl = {
     type = "object",
     properties = {
         id = id_schema,
+        type = {
+            description = "ssl certificate type, " ..
+                            "server to server certificate, " ..
+                            "client to client certificate for upstream",
+            type = "string",
+            default = "server",
+            enum = {"server", "client"}
+        },
         cert = certificate_scheme,
         key = private_key_schema,
         sni = {
@@ -763,10 +790,20 @@ _M.ssl = {
         create_time = timestamp_def,
         update_time = timestamp_def
     },
-    oneOf = {
-        {required = {"sni", "key", "cert"}},
-        {required = {"snis", "key", "cert"}}
+    ["if"] = {
+        properties = {
+            type = {
+                enum = {"server"},
+            },
+        },
     },
+    ["then"] = {
+        oneOf = {
+            {required = {"sni", "key", "cert"}},
+            {required = {"snis", "key", "cert"}}
+        }
+    },
+    ["else"] = {required = {"key", "cert"}}
 }
 
 
@@ -905,6 +942,30 @@ _M.plugin_injected_schema = {
     ["$comment"] = "this is a mark for our injected plugin schema",
     disable = {
         type = "boolean",
+    },
+    _meta = {
+        type = "object",
+        properties = {
+            error_response = {
+                oneOf = {
+                    { type = "string" },
+                    { type = "object" },
+                }
+            },
+            priority = {
+                description = "priority of plugins by customized order",
+                type = "integer",
+            },
+            filter = {
+                description = "filter determines whether the plugin "..
+                                "needs to be executed at runtime",
+                type  = "array",
+                maxItems = 20,
+                items = {
+                    type = "array"
+                }
+            }
+        }
     }
 }
 
